import numpy as np


class Dropout:
    def __init__(self, drop_rate=0.2):
        """
        Keyword Arguments:
            drop_rate {float} -- pourcentage de neurones qui ne sont pas activés
                                 à l'entrainement (default: {0.2})
        """
        self.drop_rate = drop_rate

        self.cache = None

    def forward(self, X, **kwargs):
        """Application du dropout inversé lors de la propagation avant.

        Arguments:
            X {ndarray} -- Outputs de la couche précédente.

        Keyword Arguments:
            **kwargs -- Utilisé pour indiquer si le forward
                        s'applique à l'entrainement ou au test
                        et pour inclure un seed (default: {'train', None})
        Returns:
            ndarray -- Scores de la couche
        """

        mode = kwargs.get('mode', 'train')
        seed = kwargs.get('seed', None)

        drop_mask = None

        if mode == 'train':
            # TODO
            # Remplacer la ligne A=X par la propatation avant incluant un "inverted dropout"
            # Cela se fait en 3 lignes : 
            #    1- drop_mask est un tableau binaire de la même taille que X et dont la proportion de 0 = self.drop_rate
            #    2- multiplier point-à-point X avec drop_mask
            #    3- diviser le résultat par 1-self.drop_rate. (C'est du inverted dropout) 
            #       => https://teetracker.medium.com/dropout-inverted-dropout-ec5d16d7a473
            #
            # Note : le "seed" est utilisé pour rendre déterministe la génération de nombres aléatoire.
            # C'est entre autre utile pour tester le gradients avec une fonction comme 
            # "evaluate_numerical_gradient"
            A = X
        elif mode == 'test':
            A = X
        else:
            raise Exception("Invalid forward mode %s" % mode)

        self.cache = drop_mask

        return A

    def backward(self, dA, **kwargs):
        """Rétro-propagation pour la couche de dropout inversée.

        Arguments:
            dA {ndarray} -- Gradients de la loss par rapport aux sorties.

        Keyword Arguments:
            **kwargs -- Utilisé pour indiquer si le forward
                        s'applique à l'entrainement ou au test (default: {'train'})
        Returns:
            ndarray -- Dérivée de la loss par rapport au input de la couche.
        """
        
        mode = kwargs.get('mode', 'train')

        if mode == 'train':
            drop_mask = self.cache

            dX = dA * drop_mask / (1 - self.drop_rate)
        elif mode == 'test':
            dX = dA
        else:
            raise Exception("Invalid forward mode %s" % mode)

        return dX

    def get_params(self):
        return {}

    def get_gradients(self):
        return {}

    def reset(self):
        self.__init__(drop_rate=self.drop_rate)
